@@ -1,20 +1,143 @@
+from os import environ
 import logging
 from time import sleep
 import requests
 from urllib3 import disable_warnings
 from urllib3.exceptions import InsecureRequestWarning
 
+from qbittorrentapi.decorators import login_required
 from qbittorrentapi.exceptions import *
+from qbittorrentapi.helpers import APINames
+from qbittorrentapi.helpers import suppress_context
 
 try:  # python 3
+    # noinspection PyCompatibility,PyUnresolvedReferences
     from urllib.parse import urlparse
 except ImportError:  # python 2
+    # noinspection PyCompatibility,PyUnresolvedReferences
     from urlparse import urlparse
 
 logger = logging.getLogger(__name__)
 
 
-class RequestMixIn:
+class Request(object):
+    """Facilitates HTTP requests to qBittorrent"""
+    def __init__(self, host='', port=None, username=None, password=None, **kwargs):
+        self.host = host
+        self.port = port
+        self.username = username or ''
+        self._password = password or ''
+
+        # defaults that should not change
+        self._API_URL_BASE_PATH = 'api'
+        self._API_URL_API_VERSION = 'v2'
+
+        # state, context, and caching variables
+        #   These variables are deleted if the connection to qBittorrent is reset
+        #   or a new login is required. All of these (except the SID cookie) should
+        #   be reset in _initialize_connection().
+        self._SID = None  # authorization cookie
+        self._cached_web_api_version = None
+        self._application = None
+        self._transfer = None
+        self._torrents = None
+        self._torrent_categories = None
+        self._torrent_tags = None
+        self._log = None
+        self._sync = None
+        self._rss = None
+        self._search = None
+        self._API_URL_BASE = None
+
+        # Configuration variables
+        self._VERIFY_WEBUI_CERTIFICATE = kwargs.pop('VERIFY_WEBUI_CERTIFICATE', True)
+        self._RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS = kwargs.pop(
+            'RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS',
+            kwargs.pop('RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS', False))
+        self._VERBOSE_RESPONSE_LOGGING = kwargs.pop('VERBOSE_RESPONSE_LOGGING', False)
+        self._PRINT_STACK_FOR_EACH_REQUEST = kwargs.pop('PRINT_STACK_FOR_EACH_REQUEST', False)
+        self._SIMPLE_RESPONSES = kwargs.pop('SIMPLE_RESPONSES', False)
+
+        if kwargs.pop('DISABLE_LOGGING_DEBUG_OUTPUT', False):
+            logging.getLogger('qbittorrentapi').setLevel(logging.INFO)
+            logging.getLogger('requests').setLevel(logging.INFO)
+            logging.getLogger('urllib3').setLevel(logging.INFO)
+
+        # Environment variables have lowest priority
+        if self.host == '' and environ.get('PYTHON_QBITTORRENTAPI_HOST') is not None:
+            logger.debug('Using PYTHON_QBITTORRENTAPI_HOST env variable for qBittorrent hostname')
+            self.host = environ['PYTHON_QBITTORRENTAPI_HOST']
+        if self.username == '' and environ.get('PYTHON_QBITTORRENTAPI_USERNAME') is not None:
+            logger.debug('Using PYTHON_QBITTORRENTAPI_USERNAME env variable for username')
+            self.username = environ['PYTHON_QBITTORRENTAPI_USERNAME']
+
+        if self._password == '' and environ.get('PYTHON_QBITTORRENTAPI_PASSWORD') is not None:
+            logger.debug('Using PYTHON_QBITTORRENTAPI_PASSWORD env variable for password')
+            self._password = environ['PYTHON_QBITTORRENTAPI_PASSWORD']
+
+        if self._VERIFY_WEBUI_CERTIFICATE is True and environ.get('PYTHON_QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE') is not None:
+            self._VERIFY_WEBUI_CERTIFICATE = False
+
+        # Mocking variables until better unit testing exists
+        self._MOCK_WEB_API_VERSION = kwargs.pop('MOCK_WEB_API_VERSION', None)
+
+    def _initialize_context(self):
+        """Reset context. This is necessary when the auth cookie needs to be replaced."""
+        # cache to avoid perf hit from version checking certain endpoints
+        self._cached_web_api_version = None
+
+        # reset URL so the full URL is derived again (primarily allows for switching scheme for WebUI: HTTP <-> HTTPS)
+        self._API_URL_BASE = None
+
+        # reinitialize interaction layers
+        self._application = None
+        self._transfer = None
+        self._torrents = None
+        self._torrent_categories = None
+        self._torrent_tags = None
+        self._log = None
+        self._sync = None
+        self._rss = None
+        self._search = None
+
+    @property
+    def is_logged_in(self):
+        return bool(self._SID)
+
+    def auth_log_in(self, username=None, password=None):
+        """
+        Log in to qBittorrent host.
+
+        Exceptions:
+            LoginFailed if credentials failed to log in
+            Forbidden403Error if user user is banned...or not logged in
+
+        :param username: user name for qBittorrent client
+        :param password: password for qBittorrent client
+        :return: None
+        """
+        if username:
+            self.username = username or ''
+            self._password = password or ''
+
+        try:
+            self._initialize_context()
+            response = self._post(_name=APINames.Authorization,
+                                  _method='login',
+                                  data={'username': self.username, 'password': self._password})
+            self._SID = response.cookies['SID']
+        except KeyError:
+            logger.debug('Login failed for user "%s"' % self.username)
+            raise suppress_context(LoginFailed('Login authorization failed for user "%s"' % self.username))
+        else:
+            logger.debug('Login successful for user "%s"' % self.username)
+            logger.debug('SID: %s' % self._SID)
+
+    @login_required
+    def auth_log_out(self, **kwargs):
+        """End session with qBittorrent"""
+        self._get(_name=APINames.Authorization, _method='logout', **kwargs)
+
     def _get(self, _name='', _method='', **kwargs):
         return self._request_wrapper(http_method='get', api_name=_name, api_method=_method, **kwargs)
 
@@ -22,8 +145,7 @@ class RequestMixIn:
         return self._request_wrapper(http_method='post', api_name=_name, api_method=_method, **kwargs)
 
     def _request_wrapper(self, _retries=2, _retry_backoff_factor=.3, **kwargs):
-        """Wrapper to manage requests retries."""
-
+        """ Wrapper to manage requests retries """
         # This should retry at least twice to account for the Web API switching from HTTP to HTTPS.
         # During the second attempt, the URL is rebuilt using HTTP or HTTPS as appropriate.
         max_retries = _retries if _retries > 1 else 2
@@ -39,7 +161,7 @@ class RequestMixIn:
                     error_prologue = 'Failed to connect to qBittorrent. '
                     error_messages = {
                         requests.exceptions.SSLError:
-                            'This is likely due to using an untrusted certificate (likely self-signed) ' 
+                            'This is likely due to using an untrusted certificate (likely self-signed) '
                             'for HTTPS qBittorrent WebUI. To suppress this error (and skip certificate '
                             'verification consequently exposing the HTTPS connection to man-in-the-middle '
                             'attacks), set VERIFY_WEBUI_CERTIFICATE=False when instantiating Client or set '
@@ -53,28 +175,22 @@ class RequestMixIn:
                     }
                     error_message = error_prologue + error_messages.get(type(e), 'Unknown Error: %s' % repr(e))
                     logger.debug(error_message)
-                    response = e.response if hasattr(e, 'response') else None
+                    response = getattr(e, 'response', None)
                     raise APIConnectionError(error_message, response=response)
 
             # back off on attempting each subsequent retry. first retry is always immediate.
             # if the backoff factor is 0.1, then sleep() will sleep for [0.0s, 0.2s, 0.4s, 0.8s, ...] between retries.
-            try:
-                if retry > 0:
-                    backoff_time = _retry_backoff_factor * (2 ** ((retry + 1) - 1))
-                    if backoff_time < 120:
-                        sleep(backoff_time)
-            except Exception:
-                pass
-            finally:
-                logger.debug('Retry attempt %d' % (retry+1))
-                self._initialize_context()
+            if retry > 0:
+                backoff_time = _retry_backoff_factor * (2 ** ((retry + 1) - 1))
+                sleep(backoff_time if backoff_time < 30 else 30)
+            logger.debug('Retry attempt %d' % (retry+1))
+            self._initialize_context()
 
     def _request(self, http_method, api_name, api_method,
-                 data=None, params=None, files=None, headers=None, requests_params=None,
-                 **kwargs):
+                 data=None, params=None, files=None, headers=None, requests_params=None, **kwargs):
+        _ = kwargs.pop('SIMPLE_RESPONSES', kwargs.pop('SIMPLE_RESPONSE', False))  # ensure SIMPLE_RESPONSE(S) isn't sent
 
         api_path_list = (self._API_URL_BASE_PATH, self._API_URL_API_VERSION, api_name, api_method)
-
         url = self._build_url(base_url=self._API_URL_BASE,
                               host=self.host,
                               port=self.port,
@@ -99,6 +215,10 @@ class RequestMixIn:
         headers = headers or dict()
         headers['Referer'] = self._API_URL_BASE.geturl()
         headers['Origin'] = self._API_URL_BASE.geturl()
+        # send Content-Length zero for empty POSTs
+        # Requests will not send Content-Length if data is empty
+        if http_method == 'post' and not any(filter(None, data.values())):
+            headers['Content-Length'] = '0'
 
         # include the SID auth cookie unless we're trying to log in and get a SID
         cookies = {'SID': self._SID if 'auth/login' not in url.path else ''}
@@ -117,27 +237,17 @@ class RequestMixIn:
                                     files=files,
                                     **requests_params)
 
-        if self._VERBOSE_RESPONSE_LOGGING:
-            resp_logger = logger.debug
-            max_text_length_to_log = 254
-            if response.status_code != 200:
-                max_text_length_to_log = 10000  # log as much as possible in a error condition
+        self.verbose_logging(http_method, response, url)
+        self.handle_error_responses(data, params, response)
+        return response
 
-            resp_logger('Request URL: (%s) %s' % (http_method.upper(), response.url))
-            if str(response.request.body) not in ('None', '') and 'auth/login' not in url.path:
-                body_len = max_text_length_to_log if len(response.request.body) > max_text_length_to_log else len(response.request.body)
-                resp_logger('Request body: %s%s' % (response.request.body[:body_len], '...<truncated>' if body_len >= 80 else ''))
-
-            resp_logger('Response status: %s (%s)' % (response.status_code, response.reason))
-            if response.text:
-                text_len = max_text_length_to_log if len(response.text) > max_text_length_to_log else len(response.text)
-                resp_logger('Response text: %s%s' % (response.text[:text_len], '...<truncated>' if text_len >= 80 else ''))
-
-        if self._PRINT_STACK_FOR_EACH_REQUEST:
-            from traceback import print_stack
-            print_stack()
-
-        if response.status_code == 400:
+    @staticmethod
+    def handle_error_responses(data, params, response):
+        """Raise proper exception if qBittorrent returns Error HTTP Status"""
+        if response.status_code < 400:
+            # short circuit for non-error statuses
+            return
+        elif response.status_code == 400:
             """
             Returned for malformed requests such as missing or invalid parameters.
 
@@ -200,7 +310,29 @@ class RequestMixIn:
             """
             raise HTTPError(response.text)
 
-        return response
+    def verbose_logging(self, http_method, response, url):
+        """Log verbose information about request. Can be useful during development"""
+        if self._VERBOSE_RESPONSE_LOGGING:
+            resp_logger = logger.debug
+            max_text_length_to_log = 254
+            if response.status_code != 200:
+                max_text_length_to_log = 10000  # log as much as possible in a error condition
+
+            resp_logger('Request URL: (%s) %s' % (http_method.upper(), response.url))
+            if str(response.request.body) not in ('None', '') and 'auth/login' not in url.path:
+                body_len = max_text_length_to_log if len(response.request.body) > max_text_length_to_log else len(
+                    response.request.body)
+                resp_logger('Request body: %s%s' % (
+                response.request.body[:body_len], '...<truncated>' if body_len >= 80 else ''))
+
+            resp_logger('Response status: %s (%s)' % (response.status_code, response.reason))
+            if response.text:
+                text_len = max_text_length_to_log if len(response.text) > max_text_length_to_log else len(response.text)
+                resp_logger(
+                    'Response text: %s%s' % (response.text[:text_len], '...<truncated>' if text_len >= 80 else ''))
+        if self._PRINT_STACK_FOR_EACH_REQUEST:
+            from traceback import print_stack
+            print_stack()
 
     @staticmethod
     def _build_url(base_url=None, host='', port=None, api_path_list=None):
